"use strict";
exports.__esModule = true;
var ts = require("typescript");
var _ = require("lodash");
var TsUtil = require("../util/tsnode.parser");
var CommonIslandService = /** @class */ (function () {
    function CommonIslandService(program, islandBaseService) {
        this.mongooseSchemas = {};
        this.envDecorators = {};
        this.errorEnums = [];
        this.rpcs = {};
        this.events = {};
        this.endpoints = {};
        this.crlfRegexExp = /(?:[\n\n]|[\r\n]|\\[rn])/g;
        this.program = program;
        this.islandBaseService = islandBaseService;
        if (!this.program)
            return;
        this.sourceFiles = this.program.getSourceFiles();
        this.typeChecker = this.program.getTypeChecker();
        this.parseFiles();
        // console.log(JSON.stringify(this.enumValues, null, 2));
        // console.log(JSON.stringify(this.declarations, null, 2));
    }
    CommonIslandService.prototype.parseFiles = function () {
        for (var _i = 0, _a = this.sourceFiles; _i < _a.length; _i++) {
            var sf = _a[_i];
            var fileName = sf.fileName;
            if (/node_modules/.test(fileName))
                continue;
            if (fileName.match(/(constant|environments)\.ts/)) {
                this.visit(sf);
            }
            else if (fileName.match(/(error)\.ts/)) {
                this.visit(sf);
            }
            else if (fileName.match(/(model)\.ts/)) {
                this.visit(sf);
            }
            else if (fileName.match(/(controller)\.ts/)) {
                this.visit(sf);
            }
        }
    };
    CommonIslandService.prototype.visit = function (node) {
        var _this = this;
        // console.log(TsUtil.getNodeType(node));
        var res;
        switch (node.kind) {
            case ts.SyntaxKind.VariableDeclaration:
                this.parseMongooseSchema(node);
                break;
            case ts.SyntaxKind.PropertyDeclaration:
                this.parseEnvDeclaration(node);
                break;
            case ts.SyntaxKind.EnumMember:
                // const text = TsUtil.getNodeText(node);
                // console.log(text, TsUtil.getNodeName(node.parent));
                this.parseErrorEnums(node);
                break;
            // case ts.SyntaxKind.EnumDeclaration:
            //   this.parseErrorEnums(node);
            //   break;
            case ts.SyntaxKind.MethodDeclaration:
                this.parseControllers(node);
                break;
        }
        ts.forEachChild(node, function (childNode) { return _this.visit(childNode); });
    };
    CommonIslandService.prototype.parseMongooseSchema = function (node) {
        var text = TsUtil.getNodeText(node);
        if (!/Schema/.test(text) || !node.initializer ||
            !node.initializer.arguments || !node.initializer.arguments.length) {
            return null;
        }
        var Notes = TsUtil.getNodeLeadingComments(node);
        var res = {
            Name: TsUtil.getNodeText(node.name).trim(),
            Schema: TsUtil.getNodeText(node.initializer.arguments[0]),
            ShardKey: TsUtil.getNodeText(node.initializer.arguments[1]),
            Notes: Notes
        };
        return this.mongooseSchemas[res.Name] = res;
    };
    CommonIslandService.prototype.parseEnvDeclaration = function (node) {
        var decorators = TsUtil.getNodeDecorators(node);
        if (!decorators || !decorators.some(function (d) { return /@.*env/.test(d); }))
            return null;
        var decoText = decorators.join('');
        var Notes = TsUtil.getNodeLeadingComments(node);
        var res = {
            Name: TsUtil.getNodeName(node),
            Type: TsUtil.getNodeType(node),
            Default: TsUtil.getNodeDefaultValue(node),
            Notes: Notes,
            LegacyKeys: undefined
        };
        res.LegacyKeys = TsUtil.parseEnvDecoratorText(decoText).legacyKeys;
        return this.envDecorators[res.Name] = res;
    };
    CommonIslandService.prototype.parseErrorEnums = function (node) {
        return this.errorEnums.push(TsUtil.getNodeText(node));
    };
    CommonIslandService.prototype.parseControllers = function (node) {
        return this.parseEndpointController(node)
            || this.parseRPCController(node)
            || this.parseEventController(node);
    };
    // Function Name 이 곧 RPC 의 호출 방법
    // validate query, result 를 파싱할 수 있으면 참 좋지만,,
    // param 1 번을 parsing 하게 만들자.
    // 역으로 param 1 과 return value 를 parsing 해서 decorator 를 만들어주면 사람들이 좋아할듯
    // 나 빼고..
    CommonIslandService.prototype.parseRPCController = function (node) {
        var decorators = TsUtil.getNodeDecorators(node);
        if (!decorators || !decorators.some(function (d) { return /@.*rpc/.test(d); }))
            return null;
        var Notes = TsUtil.getNodeLeadingComments(node);
        var Schemas = [];
        // Parse first parameter interface
        var params = node.parameters;
        if (params && params.length) {
            // refer to interface
            var paramText = TsUtil.getNodeText(params[0]);
            Schemas.push("Request  : " + paramText);
            if (params[0].type && params[0].type.typeName) {
                // function (params: AnotherTypeInterface)
                Schemas = Schemas.concat(this.islandBaseService.getInterfaceDeclaration(paramText));
            }
            else if (params[0].type && params[0].type.members) {
                // function (params: { type: string })
            }
            else {
                // function ({})
            }
        }
        if (node.type && node.type.typeArguments) {
            var returnTypeText = TsUtil.getNodeText(node.type.typeArguments[0]);
            Schemas.push("Response : " + returnTypeText);
            Schemas = Schemas.concat(this.islandBaseService.getInterfaceDeclaration(returnTypeText));
        }
        else {
            var returnTypeText = TsUtil.getNodeText(node.type);
            Schemas.push("Response : " + returnTypeText);
        }
        _.forEach(decorators, function (d) {
            if (!/(validate|sanitize)/.test(d))
                return;
            Schemas.push(d);
        });
        var res = {
            Name: TsUtil.getNodeName(node),
            Notes: Notes,
            Schemas: Schemas.join('\n').replace(this.crlfRegexExp, '\n').replace(/\n\s{2,3}/g, '\n')
        };
        return this.rpcs[res.Name] = res;
    };
    // @endpoint.get('/v2/aaaaa')
    // @endpoint('GET /v2/aaaaa') 
    // 두가지 문법이 혼용되어 있다.
    CommonIslandService.prototype.parseEndpointController = function (node) {
        var decorators = TsUtil.getNodeDecorators(node);
        var matched;
        if (!decorators)
            return null;
        _.some(decorators, function (d) {
            matched = d.match(/@.*(endpoint).*\((.*)\)/);
            return !!matched;
        });
        if (!matched || matched.length < 3)
            return null;
        var FuncName = TsUtil.getNodeName(node);
        var Name = matched[2] + " " + FuncName;
        var Notes = TsUtil.getNodeLeadingComments(node);
        var Schema = _.filter(decorators, function (d) {
            return /(validate|sanitize)/.test(d);
        }).join('\n').replace(this.crlfRegexExp, '\n').replace(/\n\s{2,3}/g, '\n');
        var res = {
            Name: Name,
            Notes: Notes,
            Schema: Schema
        };
        return this.endpoints[res.Name] = res;
    };
    CommonIslandService.prototype.parseEventController = function (node) {
        var _this = this;
        var decorators = TsUtil.getNodeDecorators(node);
        if (!decorators || !decorators.some(function (d) { return /@subscribe/.test(d); }))
            return null;
        var Schemas = [];
        _.forEach(decorators, function (d) {
            if (!/subscribe/.test(d))
                return;
            Schemas.push(d);
            Schemas = Schemas.concat(_this.islandBaseService.getInterfaceDeclaration(d));
            return;
        });
        var Notes = TsUtil.getNodeLeadingComments(node);
        var res = {
            Name: TsUtil.getNodeName(node),
            Notes: Notes,
            Schemas: Schemas.join('\n').replace(this.crlfRegexExp, '\n').replace(/\n\s{2,3}/g, '\n')
        };
        return this.events[res.Name] = res;
    };
    return CommonIslandService;
}());
exports.CommonIslandService = CommonIslandService;
