// tslint:disable-next-line
require('source-map-support').install();

// https://github.com/Microsoft/TypeScript/wiki/Using-the-Compiler-API
// https://github.com/basarat/typescript-book/blob/master/docs/compiler/ast-trivia.md
// https://github.com/Microsoft/TypeScript/blob/master/src/compiler/types.ts
// See typescript.d.ts
import * as ts from 'typescript';
import * as _ from 'lodash';
import * as fs from 'fs';
import * as glob from 'glob';

import * as tsnodeParser from './util/tsnode.parser';
import * as mdExporter from './util/md.exporter';
import { ConfluenceUploader } from './util/confluence.uploader';

import { IslandBaseService } from './service/island.base.service';
import { CommonIslandService } from './service/common.island.service';

class islandDoc {
  options: any = {
    mdFilePath: './README.md',
    mdFilesuffix: '',
    sourceFilePath: './src/',
    sourceFilePattern: '+(*controller|environments|constant|error|*model).ts',
    globOptions: {
      ignore: ["**/node_modules*/**", "./node_modules/**"]
    }
  };

  constructor() {}
  
  getProgram(path: string, pattern: string = this.options.sourceFilePattern, includeNodeModules: boolean = false): ts.Program {
    if (!fs.existsSync(path)) return null;
    const files = glob.sync(
      `${path}/**/${pattern}`,
      { ignore: includeNodeModules ? [] : ["**/node_modules/**", "./node_modules/**"] }
    );
    if (files.length < 1) return null;

    const tsconfig = JSON.parse(fs.readFileSync('./tsconfig.json').toString('utf-8'));
    const compilerOptions: any = ts.convertCompilerOptionsFromJson(tsconfig.compilerOptions, '.');

    const program: ts.Program = ts.createProgram(files, compilerOptions.options);
    // const checker: ts.TypeChecker = program.getTypeChecker();
    return program;
  }

  async run(callback?: Function, options?) {
    try {
      const testMode: boolean = options && options.test;

      if (testMode) {
        this.options.mdFilePath = './spec/*.md';
        this.options.mdFilesuffix = '.md';
        this.options.sourceFilePath = './spec/';
        this.options.sourceFilePattern = `sample*.ts`;
      }

      const ibsProgram = this.getProgram('./node_modules/island-base/src/', '*.ts', true);
      const ibs = new IslandBaseService(ibsProgram);

      const program: ts.Program = this.getProgram(this.options.sourceFilePath, this.options.sourceFilePattern);
      const cis = new CommonIslandService(program, ibs);

      const cu = new ConfluenceUploader();

      const envText = mdExporter.createTableFromMdText(cis.envDecorators);
      const mongooseText = mdExporter.createMdText(cis.mongooseSchemas);
      const islandCodes = ibs.getValue('IslandCode'); 
      const errorEnums = cis.errorEnums;
      const errorText = mdExporter.createMdText(islandCodes.concat(errorEnums));

      if (envText || mongooseText || errorText) {
        let readMeContent = fs.existsSync('./README.md') ? fs.readFileSync('./README.md', 'utf8') : '';
        
        readMeContent = mdExporter.replaceSection(readMeContent, 'Error Codes', errorText);
        readMeContent = mdExporter.replaceSection(readMeContent, 'Environment Variables', envText);
        readMeContent = mdExporter.replaceSection(readMeContent, 'Mongoose Schemas', mongooseText);

        const saveFileName = `./README.md${this.options.mdFilesuffix}`;
        fs.writeFileSync(saveFileName, readMeContent, 'utf8');

        if (options && options.updateConfluence) {
          console.log(1);
          await cu.uploadErrorCodes(this.getProjectInfo(), errorEnums, islandCodes);
        }
      }

      const eventText = mdExporter.createMdText(cis.events);
      if (eventText) {
        let eventContent = fs.existsSync('./EVENT.md') ? fs.readFileSync('./EVENT.md', 'utf8') : '';
        const saveFileName = `./EVENT.md${this.options.mdFilesuffix}`;
        fs.writeFileSync(saveFileName, eventContent, 'utf8');
        eventContent = mdExporter.replaceSection(eventContent, 'Event Schemas', eventText);
      }

      const endpointText = mdExporter.createMdText(cis.endpoints);
      if (endpointText) {
        let endpointContent = fs.existsSync('./ENDPOINT.md') ? fs.readFileSync('./ENDPOINT.md', 'utf8') : '';
        endpointContent = mdExporter.replaceSection(endpointContent, 'Endpoint Schemas', endpointText);
        const saveFileName = `./ENDPOINT.md${this.options.mdFilesuffix}`;
        fs.writeFileSync(saveFileName, endpointContent, 'utf8');
      }

      const rpcText = mdExporter.createMdText(cis.rpcs);
      if (rpcText) {
        let rpcContent = fs.existsSync('./RPC.md') ? fs.readFileSync('./RPC.md', 'utf8') : '';
        rpcContent = mdExporter.replaceSection(rpcContent, 'RPC Schemas', rpcText);
        const saveFileName = `./RPC.md${this.options.mdFilesuffix}`;
        fs.writeFileSync(saveFileName, rpcContent, 'utf8');
      }
    } catch (err) {
      console.error(err);
    } finally {
      return typeof callback === 'function' ? callback() : undefined;
    }
  }

  getProjectInfo() {
    try {
      const pkgJson = require(`${process.cwd()}/package.json`);
      return {
        name: pkgJson.name,
        version: pkgJson.version
      };
    } catch (e) {
      return {};
    }
  }
}

const islandDocMaker = new islandDoc();
export default islandDocMaker;
