import * as ts from 'typescript';
import * as _ from 'lodash';
import { IslandBaseService } from '../service/island.base.service';
import * as TsUtil from '../util/tsnode.parser';

export class CommonIslandService {
  private program: ts.Program;
  private islandBaseService: IslandBaseService;
  private typeChecker: ts.TypeChecker;
  private sourceFiles: ReadonlyArray<ts.SourceFile>;

  public mongooseSchemas: any = {};
  public envDecorators: any = {};
  public errorEnums: any = [];
  public rpcs: any = {};
  public events: any = {};
  public endpoints: any = {};

  private crlfRegexExp: RegExp = /(?:[\n\n]|[\r\n]|\\[rn])/g;

  constructor(program: ts.Program, islandBaseService: IslandBaseService) {
    this.program = program;
    this.islandBaseService = islandBaseService;
    if (!this.program) return;

    this.sourceFiles = this.program.getSourceFiles();
    this.typeChecker = this.program.getTypeChecker();

    this.parseFiles();

    // console.log(JSON.stringify(this.enumValues, null, 2));
    // console.log(JSON.stringify(this.declarations, null, 2));
  }

  private parseFiles() {
    for (const sf of this.sourceFiles) {
      const fileName = sf.fileName;

      if (/node_modules/.test(fileName)) continue;
      
      if (fileName.match(/(constant|environments)\.ts/)) {
        this.visit(sf);
      } else if (fileName.match(/(error)\.ts/)) {
        this.visit(sf);
      } else if (fileName.match(/(model)\.ts/)) {
        this.visit(sf);
      } else if (fileName.match(/(controller)\.ts/)) {
        this.visit(sf);
      }
    }
  }

  private visit(node: ts.Node) {
    // console.log(TsUtil.getNodeType(node));
    let res;
    switch (node.kind) {
      case ts.SyntaxKind.VariableDeclaration:
        this.parseMongooseSchema(node);
        break;
      case ts.SyntaxKind.PropertyDeclaration:
        this.parseEnvDeclaration(node);
        break;
      case ts.SyntaxKind.EnumMember:
        // const text = TsUtil.getNodeText(node);
        // console.log(text, TsUtil.getNodeName(node.parent));
        this.parseErrorEnums(node);
        break;
      // case ts.SyntaxKind.EnumDeclaration:
      //   this.parseErrorEnums(node);
      //   break;
      case ts.SyntaxKind.MethodDeclaration:
        this.parseControllers(node);
        break;
    }

    ts.forEachChild(node, (childNode) => this.visit(childNode));
  }

  private parseMongooseSchema(node: any) {
    const text = TsUtil.getNodeText(node);
    if (!/Schema/.test(text) || !node.initializer || 
        !node.initializer.arguments || !node.initializer.arguments.length) {
          return null;
        }

    const Notes = TsUtil.getNodeLeadingComments(node);
    const res = {
      Name: TsUtil.getNodeText(node.name).trim(),
      Schema: TsUtil.getNodeText(node.initializer.arguments[0]),
      ShardKey: TsUtil.getNodeText(node.initializer.arguments[1]),
      Notes
    };

    return this.mongooseSchemas[res.Name] = res;
  }

  private parseEnvDeclaration(node: any) {
    const decorators: string[] = TsUtil.getNodeDecorators(node);
    if (!decorators || !decorators.some((d) => { return /@.*env/.test(d); })) return null;

    const decoText: string = decorators.join('');
    const Notes = TsUtil.getNodeLeadingComments(node);

    const res = {
      Name: TsUtil.getNodeName(node),
      Type: TsUtil.getNodeType(node),
      Default: TsUtil.getNodeDefaultValue(node),
      Notes,
      LegacyKeys: undefined,
    };

    res.LegacyKeys = TsUtil.parseEnvDecoratorText(decoText).legacyKeys;
    return this.envDecorators[res.Name] = res;
  }

  private parseErrorEnums(node: any) {
    return this.errorEnums.push(TsUtil.getNodeText(node));
  }

  private parseControllers(node: any) {
    return this.parseEndpointController(node)
    || this.parseRPCController(node)
    || this.parseEventController(node);
  }

  // Function Name 이 곧 RPC 의 호출 방법
  // validate query, result 를 파싱할 수 있으면 참 좋지만,,
  // param 1 번을 parsing 하게 만들자.
  // 역으로 param 1 과 return value 를 parsing 해서 decorator 를 만들어주면 사람들이 좋아할듯
  // 나 빼고..
  private parseRPCController(node: any) : any {
    const decorators: string[] = TsUtil.getNodeDecorators(node);
    if (!decorators || !decorators.some((d) => { return /@.*rpc/.test(d); })) return null;

    const Notes = TsUtil.getNodeLeadingComments(node);
    let Schemas = [];
    // Parse first parameter interface
    const params = node.parameters;
    if (params && params.length) {
      // refer to interface
      const paramText = TsUtil.getNodeText(params[0]);
      Schemas.push(`Request  : ${paramText}`);
      if (params[0].type && params[0].type.typeName) {
        // function (params: AnotherTypeInterface)
        Schemas = Schemas.concat(this.islandBaseService.getInterfaceDeclaration(paramText));
      } else if (params[0].type && params[0].type.members) {
        // function (params: { type: string })
      } else {
        // function ({})
      }
    }

    if (node.type && node.type.typeArguments) {
      const returnTypeText = TsUtil.getNodeText(node.type.typeArguments[0]);
      Schemas.push(`Response : ${returnTypeText}`);
      Schemas = Schemas.concat(this.islandBaseService.getInterfaceDeclaration(returnTypeText));
    } else {
      const returnTypeText = TsUtil.getNodeText(node.type);
      Schemas.push(`Response : ${returnTypeText}`);
    }

    _.forEach(decorators, (d) => {
      if (!/(validate|sanitize)/.test(d)) return;
      Schemas.push(d);
    });
    const res = {
      Name: TsUtil.getNodeName(node),
      Notes,
      Schemas: Schemas.join('\n').replace(this.crlfRegexExp, '\n').replace(/\n\s{2,3}/g, '\n')
    };

    return this.rpcs[res.Name] = res;
  }

  // @endpoint.get('/v2/aaaaa')
  // @endpoint('GET /v2/aaaaa') 
  // 두가지 문법이 혼용되어 있다.
  private parseEndpointController(node: any) : any {
    const decorators: string[] = TsUtil.getNodeDecorators(node);
    let matched;

    if (!decorators) return null;
    _.some(decorators, (d) => {
      matched = d.match(/@.*endpoint(\..*)?\((.*)\)/);
      return !!matched;
    });

    if (!matched || matched.length < 3) return null;

    const FuncName = TsUtil.getNodeName(node);
    const Name = matched[1] ? `${matched[1]} ${matched[2]} ${FuncName}` : `${matched[2]} ${FuncName}`;
    const Notes = TsUtil.getNodeLeadingComments(node);
    const Schema = _.filter(decorators, (d) => {
      return /(validate|sanitize)/.test(d);
    }).join('\n').replace(this.crlfRegexExp, '\n').replace(/\n\s{2,3}/g, '\n');

    const res = {
      Name,
      Notes,
      Schema
    };

    return this.endpoints[res.Name] = res;
  }

  private parseEventController(node: any) : any {
    const decorators: string[] = TsUtil.getNodeDecorators(node);
    if (!decorators || !decorators.some((d) => { return /@subscribe/.test(d); })) return null;

    let Schemas = [];
    _.forEach(decorators, (d) => {
      if (!/subscribe/.test(d)) return;

      Schemas.push(d);
      Schemas = Schemas.concat(this.islandBaseService.getInterfaceDeclaration(d));
      return;
    });

    const Notes = TsUtil.getNodeLeadingComments(node);
    const res = {
      Name: TsUtil.getNodeName(node),
      Notes,
      Schemas: Schemas.join('\n').replace(this.crlfRegexExp, '\n').replace(/\n\s{2,3}/g, '\n')
    };

    return this.events[res.Name] = res;
  }
}