import * as _ from 'lodash';
import * as path from 'path';
import * as ts from 'typescript';

export const getSourceFile = (anyTsObject: any): any => {
  if (anyTsObject.parent) return getSourceFile(anyTsObject.parent);
  if (Array.isArray(anyTsObject) && anyTsObject.length) {
    return getSourceFile(anyTsObject[0]);
  }
  return anyTsObject;
};

export const findNode = (node: ts.Node, kind: ts.SyntaxKind) => {
  const found = [];
  ts.forEachChild(node, childNode => {
    if (!kind || childNode.kind === kind) {
      found.push(childNode);
    }
  });
  return found;
};

export const getNodeName = (node: ts.Node): string | undefined => {
  if (!node) return undefined;

  const anyNode = node as any;
  return anyNode.escapedName 
        || (anyNode.name && anyNode.name.escapedText)
        || undefined;
};

export const getDeepName = (node: ts.Node, names = []): String[] => {
  // if there is no parent, that node is source file
  if (node.parent) {
    const nodeName = getNodeName(node);
    if (nodeName) { names.unshift(nodeName); }
    getDeepName(node.parent, names);
  } else {
    const nodeName = path.basename((node as any).path);
    if (nodeName) {
      names.unshift(nodeName.replace(/\.ts/, ''));
    }
  }

  return names;
};

export const getNodeType = (node: ts.Node) => {
  const type = (node as any).type || node;
  if (!type || !type.kind) {
    console.warn(`some node has no type declaration. treat it as any`);
    return 'Treat It As Any';
  }
  
  return ts.SyntaxKind[type.kind].replace('Keyword', '');
};

export const getNodeText = (node: any): string => {
  if (!node || node.pos === undefined || node.end === undefined) return '';
  const sourceFile: any = getSourceFile(node);
  
  if (node.getText) return node.getText(sourceFile);

  const text = sourceFile.getText();
  return text.substring(node.pos, node.end);
};


export const getNodeLeadingComments = (node: ts.Node): string => {
  const sourceFile: ts.SourceFile = node.getSourceFile();
  const comments = ts.getLeadingCommentRanges(sourceFile.getText(), node.getFullStart());
  if (!comments) {
    return '';
  }

  const texts: string[] = [];
  for (const comment of comments) {
    texts.push(sourceFile.getText().substring(comment.pos, comment.end))
  }

  return texts.length ? texts.join(' ') : '';
};

export const getNodeDecorators = (node: any): string[] | undefined => {
  const decorators = node.decorators;
  if (!decorators) {
    return undefined;
  }

  const texts: string[] = [];
  for (const decorator of decorators) {
    texts.push(getNodeText(decorator));
  }

  return texts.length ? texts : undefined;
};

export const getNodeDefaultValue = (node: ts.Node): string | undefined => {
  const text = getNodeText(node);
  const matched = text.match(/\=\s*(.+)\;/);

  return matched ? matched[1] : undefined;
};

export const parseEnvDecoratorText = (text: string) => {
  const legacyKeys = text.match(/legacyKeys\:\s*\[([A-z\_\s\,\'\"]+)\]/);
  return { 
    legacyKeys: legacyKeys ? legacyKeys[1] : undefined,
    required: false
  };
};
